@@ -32,7 +32,7 @@ class Coverage(object):
                       self.wrapped,
                       total))
         else:
-            fd.write("***INFO*** There are no declared %s." % self.name)
+            fd.write("***INFO*** There are no declared %s.\n" % self.name)
 
 functions_coverage = Coverage("global functions")
 methods_coverage = Coverage("methods")
@@ -91,7 +91,7 @@ class FileOutput:
 
 class Wrapper:
     type_tmpl = (
-        'PyTypeObject Py%(typename)s_Type = {\n'
+        'PyTypeObject G_GNUC_INTERNAL Py%(typename)s_Type = {\n'
         '    PyObject_HEAD_INIT(NULL)\n'
         '    0,                                 /* ob_size */\n'
         '    "%(classname)s",                   /* tp_name */\n'
@@ -723,10 +723,109 @@ class Wrapper:
 
         return getsets_name
 
-    def write_functions(self, prefix):
-        self.fp.write('\n/* ----------- functions ----------- */\n\n')
-        functions = []
+    def _write_get_symbol_names(self, writer, functions):
+        self.fp.write("""static PyObject *
+_wrap__get_symbol_names(PyObject *self)
+{
+    PyObject *pylist = PyList_New(0);
 
+""")
+        for obj, bases in writer.get_classes():
+            self.fp.write('    PyList_Append(pylist, '
+                          'PyString_FromString("%s"));\n' % (obj.name))
+
+        for name, cname, flags, docstring in functions:
+            self.fp.write('    PyList_Append(pylist, '
+                          'PyString_FromString("%s"));\n' % (name))
+
+        for enum in writer.get_enums():
+            self.fp.write('    PyList_Append(pylist, '
+                          'PyString_FromString("%s"));\n' % (enum.name))
+            for nick, value in enum.values:
+                name = value[len(self.overrides.modulename)+1:]
+                self.fp.write('    PyList_Append(pylist, '
+                              'PyString_FromString("%s"));\n' % (name))
+
+        self.fp.write("    return pylist;\n}\n\n");
+
+    def _write_get_symbol(self, writer, functions):
+        self.fp.write("""static PyObject *
+_wrap__get_symbol(PyObject *self, PyObject *args)
+{
+    PyObject *d;
+    char *name;
+    static PyObject *modulename = NULL;
+    static PyObject *module = NULL;
+    static char *strip_prefix = "%s";
+
+    if (!PyArg_ParseTuple(args, "Os", &d, &name))
+        return NULL;
+
+    if (!modulename)
+       modulename = PyString_FromString("%s");
+
+    if (!module)
+       module = PyDict_GetItemString(d, "__module__");
+
+""" % (self.overrides.modulename.upper() + '_',
+       self.overrides.modulename))
+
+        first = True
+        # Classes / GObjects
+        for obj, bases in writer.get_classes():
+            if first:
+                self.fp.write('    if (!strcmp(name, "%s")) {\n' % obj.name)
+                first = False
+            else:
+                self.fp.write('    } else if (!strcmp(name, "%s")) {\n' % obj.name)
+            self.fp.write(
+                '       return (PyObject*)pygobject_lookup_class(%s);\n' %
+                obj.typecode)
+        self.fp.write('    }\n')
+
+        # Functions
+        for name, cname, flags, docstring in functions:
+            self.fp.write('    else if (!strcmp(name, "%s")) {\n' % name)
+            self.fp.write('        static PyMethodDef ml = { '
+                          '"%s", (PyCFunction)%s, %s, "%s"};\n' % (
+                name, cname, flags, docstring))
+            self.fp.write('        return PyCFunction_NewEx(&ml, NULL, modulename);\n')
+            self.fp.write('    }\n')
+
+        # Enums
+        def write_enum(enum, returnobj=False):
+            if returnobj:
+                ret = 'return '
+            else:
+                ret = ''
+            if enum.deftype == 'enum':
+                self.fp.write(
+                    '        %spyg_enum_add(module, "%s", strip_prefix, %s);\n'
+                    % (ret, enum.name, enum.typecode))
+            else:
+                self.fp.write(
+                    '    %spyg_flags_add(module, "%s", strip_prefix, %s);\n'
+                    % (ret, enum.name, enum.typecode))
+
+        strip_len = len(self.overrides.modulename)+1 # GTK_
+        for enum in writer.get_enums():
+            # XXX: Implement without typecodes
+            self.fp.write('    else if (!strcmp(name, "%s")) {\n' % enum.name)
+            write_enum(enum, returnobj=True)
+            self.fp.write('    }\n')
+
+            for nick, value in enum.values:
+                value = value[strip_len:]
+                self.fp.write('    else if (!strcmp(name, "%s")) {\n' % value)
+                write_enum(enum)
+                self.fp.write('        return PyObject_GetAttrString(module, "%s");\n' %
+                              value)
+                self.fp.write('    }\n')
+
+        self.fp.write('    return Py_None;\n}\n\n');
+
+    def _write_function_bodies(self):
+        functions = []
         # First, get methods from the defs files
         for func in self.parser.find_functions():
             funcname = func.c_name
@@ -743,11 +842,8 @@ class Wrapper:
                     code, methflags = self.write_function_wrapper(func,
                         self.function_tmpl, handle_return=1, is_method=0)
                     self.fp.write(code)
-                functions.append(self.methdef_tmpl %
-                                 { 'name':  func.name,
-                                   'cname': '_wrap_' + funcname,
-                                   'flags': methflags,
-                                   'docstring': func.docstring })
+                functions.append((func.name, '_wrap_' + funcname,
+                                  methflags, func.docstring))
                 functions_coverage.declare_wrapped()
             except:
                 functions_coverage.declare_not_wrapped()
@@ -760,16 +856,49 @@ class Wrapper:
                 data = self.overrides.function(funcname)
                 self.write_function(funcname, data)
                 methflags = self.get_methflags(funcname)
-                functions.append(self.methdef_tmpl %
-                                 { 'name':  funcname,
-                                   'cname': '_wrap_' + funcname,
-                                   'flags': methflags,
-                                   'docstring': 'NULL'})
+                functions.append((funcname, '_wrap_' + funcname,
+                                  methflags, 'NULL'))
                 functions_coverage.declare_wrapped()
             except:
                 functions_coverage.declare_not_wrapped()
                 sys.stderr.write('Could not write function %s: %s\n'
                                  % (funcname, exc_info()))
+        return functions
+
+    def write_functions(self, writer, prefix):
+        self.fp.write('\n/* ----------- functions ----------- */\n\n')
+        functions = []
+        func_infos = self._write_function_bodies()
+
+        # If we have a dynamic namespace, write symbol and attribute getter
+        if self.overrides.dynamicnamespace:
+            self._write_get_symbol_names(writer, func_infos)
+            self._write_get_symbol(writer, func_infos)
+            for obj, bases in writer.get_classes():
+                self.fp.write("""static PyTypeObject *
+%s_register_type(const gchar *name, PyObject *unused)
+{
+    PyObject *m = PyImport_ImportModule("gtk");
+    PyObject *d = PyModule_GetDict(m);
+""" % obj.c_name)
+                writer.write_class(obj, bases, indent=1)
+                self.fp.write(
+                    '    return (%s)PyDict_GetItemString(d, "%s");\n' % (
+                    'PyTypeObject*', obj.name))
+                self.fp.write("}\n")
+
+            functions.append('    { "_get_symbol_names", '
+                             '(PyCFunction)_wrap__get_symbol_names, '
+                             'METH_NOARGS, NULL },\n')
+            functions.append('    { "_get_symbol", '
+                             '(PyCFunction)_wrap__get_symbol, '
+                             'METH_VARARGS, NULL },\n')
+        else:
+            for name, cname, flags, docstring in func_infos:
+                functions.append(self.methdef_tmpl % dict(name=name,
+                                                          cname=cname,
+                                                          flags=flags,
+                                                          docstring=docstring))
 
         # write the PyMethodDef structure
         functions.append('    { NULL, NULL, 0, NULL }\n')
@@ -1047,7 +1176,7 @@ class GInterfaceWrapper(GObjectWrapper):
                 sys.stderr.write('Could not write interface proxy %s.%s: %s\n'
                                 % (klass, meth.name, exc_info()))
 
-        if not proxies:
+        if not proxies or not [cname for name, cname in proxies if cname]:
             return
 
         ## Write an interface init function for this object
@@ -1207,12 +1336,12 @@ class SourceWriter:
         self.write_classes()
 
         wrapper = Wrapper(self.parser, None, self.overrides, self.fp)
-        wrapper.write_functions(self.prefix)
+        wrapper.write_functions(self, self.prefix)
 
-        self.write_enums()
         if not self.overrides.dynamicnamespace:
-            self.write_extension_init()
-            self.write_registers()
+            self.write_enums()
+        self.write_extension_init()
+        self.write_registers()
 
     def write_headers(self):
         self.fp.write('/* -- THIS FILE IS GENERATED - DO NOT EDIT */')
@@ -1230,13 +1359,17 @@ class SourceWriter:
         self.fp.write('\n\n')
 
     def write_type_declarations(self):
+        #todo use 'static' if used only in one file
         self.fp.write('/* ---------- forward type declarations ---------- */\n')
         for obj in self.parser.boxes:
-            self.fp.write('PyTypeObject Py' + obj.c_name + '_Type;\n')
+            if not self.overrides.is_type_ignored(obj.c_name):
+                self.fp.write('PyTypeObject G_GNUC_INTERNAL Py' + obj.c_name + '_Type;\n')
         for obj in self.parser.objects:
-            self.fp.write('PyTypeObject Py' + obj.c_name + '_Type;\n')
+            if not self.overrides.is_type_ignored(obj.c_name):
+                self.fp.write('PyTypeObject G_GNUC_INTERNAL Py' + obj.c_name + '_Type;\n')
         for interface in self.parser.interfaces:
-            self.fp.write('PyTypeObject Py' + interface.c_name + '_Type;\n')
+            if not self.overrides.is_type_ignored(interface.c_name):
+                self.fp.write('PyTypeObject G_GNUC_INTERNAL Py' + interface.c_name + '_Type;\n')
         self.fp.write('\n')
 
     def write_body(self):
@@ -1283,17 +1416,24 @@ class SourceWriter:
                 instance.write_class()
                 self.fp.write('\n')
 
+    def get_enums(self):
+        enums = []
+        for enum in self.parser.enums:
+            if self.overrides.is_type_ignored(enum.c_name):
+                continue
+            enums.append(enum)
+        return enums
+
     def write_enums(self):
         if not self.parser.enums:
             return
+
         self.fp.write('\n/* ----------- enums and flags ----------- */\n\n')
         self.fp.write(
             'void\n' + self.prefix +
             '_add_constants(PyObject *module, const gchar *strip_prefix)\n{\n')
 
-        for enum in self.parser.enums:
-            if self.overrides.is_type_ignored(enum.c_name):
-                continue
+        for enum in self.get_enums():
             if enum.typecode is None:
                 for nick, value in enum.values:
                     self.fp.write(
@@ -1313,42 +1453,49 @@ class SourceWriter:
         self.fp.write('    PyErr_Print();\n')
         self.fp.write('}\n\n')
 
+    def write_object_imports(self, retval=''):
+        imports = self.overrides.get_imports()[:]
+        if not imports:
+            return
+
+        bymod = {}
+        for module, pyname, cname in imports:
+            bymod.setdefault(module, []).append((pyname, cname))
+        self.fp.write('    PyObject *module;\n\n')
+        for module in bymod:
+            self.fp.write(
+                '    if ((module = PyImport_ImportModule("%s")) != NULL) {\n'
+                % module)
+            #self.fp.write(
+            #    '        PyObject *moddict = PyModule_GetDict(module);\n\n')
+            for pyname, cname in bymod[module]:
+                #self.fp.write(
+                #    '        _%s = (PyTypeObject *)PyDict_GetItemString('
+                #    'moddict, "%s");\n' % (cname, pyname))
+                self.fp.write(
+                    '        _%s = (PyTypeObject *)PyObject_GetAttrString('
+                    'module, "%s");\n' % (cname, pyname))
+                self.fp.write('        if (_%s == NULL) {\n' % cname)
+                self.fp.write('            PyErr_SetString(PyExc_ImportError,\n')
+                self.fp.write('                "cannot import name %s from %s");\n'
+                         % (pyname, module))
+                self.fp.write('            return %s;\n' % retval)
+                self.fp.write('        }\n')
+            self.fp.write('    } else {\n')
+            self.fp.write('        PyErr_SetString(PyExc_ImportError,\n')
+            self.fp.write('            "could not import %s");\n' % module)
+            self.fp.write('        return %s;\n' % retval)
+            self.fp.write('    }\n')
+        self.fp.write('\n')
+
     def write_extension_init(self):
         self.fp.write('/* initialise stuff extension classes */\n')
         self.fp.write('void\n' + self.prefix + '_register_classes(PyObject *d)\n{\n')
-        imports = self.overrides.get_imports()[:]
-        if imports:
-            bymod = {}
-            for module, pyname, cname in imports:
-                bymod.setdefault(module, []).append((pyname, cname))
-            self.fp.write('    PyObject *module;\n\n')
-            for module in bymod:
-                self.fp.write(
-                    '    if ((module = PyImport_ImportModule("%s")) != NULL) {\n'
-                    % module)
-                self.fp.write(
-                    '        PyObject *moddict = PyModule_GetDict(module);\n\n')
-                for pyname, cname in bymod[module]:
-                    self.fp.write(
-                        '        _%s = (PyTypeObject *)PyDict_GetItemString('
-                        'moddict, "%s");\n' % (cname, pyname))
-                    self.fp.write('        if (_%s == NULL) {\n' % cname)
-                    self.fp.write('            PyErr_SetString(PyExc_ImportError,\n')
-                    self.fp.write('                "cannot import name %s from %s");\n'
-                             % (pyname, module))
-                    self.fp.write('            return;\n')
-                    self.fp.write('        }\n')
-                self.fp.write('    } else {\n')
-                self.fp.write('        PyErr_SetString(PyExc_ImportError,\n')
-                self.fp.write('            "could not import %s");\n' % module)
-                self.fp.write('        return;\n')
-                self.fp.write('    }\n')
-            self.fp.write('\n')
+        self.write_object_imports()
         self.fp.write(self.overrides.get_init() + '\n')
         self.fp.resetline()
 
-
-    def _get_classes(self):
+    def get_classes(self):
         objects = self.parser.objects[:]
         pos = 0
         while pos < len(objects):
@@ -1375,40 +1522,65 @@ class SourceWriter:
 
     def write_registers(self):
         for boxed in self.parser.boxes:
-            self.fp.write('    pyg_register_boxed(d, "' + boxed.name +
-                          '", ' + boxed.typecode +
-                     ', &Py' + boxed.c_name +
+            if not self.overrides.is_type_ignored(boxed.c_name):
+                self.fp.write('    pyg_register_boxed(d, "' + boxed.name +
+                              '", ' + boxed.typecode +
+                              ', &Py' + boxed.c_name +
                           '_Type);\n')
         for pointer in self.parser.pointers:
-            self.fp.write('    pyg_register_pointer(d, "' + pointer.name +
-                          '", ' + pointer.typecode +
-                          ', &Py' + pointer.c_name + '_Type);\n')
+            if not self.overrides.is_type_ignored(pointer.c_name):
+                self.fp.write('    pyg_register_pointer(d, "' + pointer.name +
+                              '", ' + pointer.typecode +
+                              ', &Py' + pointer.c_name + '_Type);\n')
         for interface in self.parser.interfaces:
-            self.fp.write('    pyg_register_interface(d, "' + interface.name +
-                     '", '+ interface.typecode + ', &Py' + interface.c_name +
-                     '_Type);\n')
-            if interface.interface_info is not None:
-                self.fp.write('    pyg_register_interface_info(%s, &%s);\n' %
-                              (interface.typecode, interface.interface_info))
+            if not self.overrides.is_type_ignored(interface.c_name):
+                self.fp.write('    pyg_register_interface(d, "'
+                              + interface.name + '", '+ interface.typecode
+                              + ', &Py' + interface.c_name + '_Type);\n')
+                if interface.interface_info is not None:
+                    self.fp.write('    pyg_register_interface_info(%s, &%s);\n' %
+                                  (interface.typecode, interface.interface_info))
+
+        if not self.overrides.dynamicnamespace:
+            for obj, bases in self.get_classes():
+                self.write_class(obj, bases)
+        else:
+            for obj, bases in self.get_classes():
+                self.fp.write(
+                    '    pyg_type_register_custom_callback("%s", '
+                    '(PyGTypeRegistrationFunction)%s_register_type, d);\n' %
+                    (obj.c_name, obj.c_name))
 
-        for obj, bases in self._get_classes():
-            self._write_class(obj, bases)
         self.fp.write('}\n')
 
-    def _write_class(self, obj, bases, indent=1):
+    def _can_direct_ref(self, base):
+        if not self.overrides.dynamicnamespace:
+            return True
+        if base == 'GObject':
+            return True
+        obj = get_object_by_name(base)
+        if obj.module.lower() != self.overrides.modulename:
+            return True
+        return False
+
+    def write_class(self, obj, bases, indent=1):
         indent_str = ' ' * (indent * 4)
         if bases:
-            bases_str = (
-                'Py_BuildValue("(' + 'O' * len(bases) + ')", ' +
-                string.join(map(lambda s: '&Py'+s+'_Type', bases), ', ') +
-                ')')
+            bases_str = 'Py_BuildValue("(%s)"' % (len(bases) * 'O')
+
+            for base in bases:
+                if self._can_direct_ref(base):
+                    bases_str += ', &Py%s_Type' % base
+                else:
+                    baseobj = get_object_by_name(base)
+                    bases_str += ', PyObject_GetAttrString(m, "%s")' % baseobj.name
+            bases_str += ')'
         else:
             bases_str = 'NULL'
 
         self.fp.write(
-                indent_str + 'pygobject_register_class(d, "' + obj.c_name +
-                '", ' + obj.typecode + ', &Py' + obj.c_name +
-                '_Type, ' + bases_str + ');\n')
+                '%(indent)spygobject_register_class(d, "%(c_name)s", %(typecode)s, &Py%(c_name)s_Type, %(bases)s);\n'
+                % dict(indent=indent_str, c_name=obj.c_name, typecode=obj.typecode, bases=bases_str))
 
         if obj.has_new_constructor_api:
             self.fp.write(
@@ -1425,15 +1597,25 @@ class SourceWriter:
                 indent_str + 'pyg_register_class_init(%s, %s);\n' %
                 (obj.typecode, obj.class_init_func))
 
+_objects = {}
+
+def get_object_by_name(c_name):
+    global _objects
+    return _objects[c_name]
+
 def register_types(parser):
+    global _objects
     for boxed in parser.boxes:
         argtypes.matcher.register_boxed(boxed.c_name, boxed.typecode)
+        _objects[boxed.c_name] = boxed
     for pointer in parser.pointers:
         argtypes.matcher.register_pointer(pointer.c_name, pointer.typecode)
     for obj in parser.objects:
         argtypes.matcher.register_object(obj.c_name, obj.parent, obj.typecode)
-    for obj in parser.interfaces:
-        argtypes.matcher.register_object(obj.c_name, None, obj.typecode)
+        _objects[obj.c_name] = obj
+    for iface in parser.interfaces:
+        argtypes.matcher.register_object(iface.c_name, None, iface.typecode)
+        _objects[iface.c_name] = iface
     for enum in parser.enums:
         if enum.deftype == 'flags':
             argtypes.matcher.register_flag(enum.c_name, enum.typecode)
