@@ -394,10 +394,12 @@ del ctype
 class StringReturn(ReturnType):
 
     def get_c_type(self):
-        return "char *"
+        return self.props.get('c_type', 'char *').replace('const-', 'const ')
+    #return "char *"
 
     def write_decl(self):
-        self.wrapper.add_declaration("char *retval;")
+        self.wrapper.add_declaration("%s retval;" % self.get_c_type())
+        #self.wrapper.add_declaration("char *retval;")
 
     def write_error_return(self):
         self.wrapper.write_code("return NULL;")
@@ -406,7 +408,7 @@ class StringReturn(ReturnType):
         self.wrapper.add_pyret_parse_item("s", "&retval", prepend=True)
         self.wrapper.write_code("retval = g_strdup(retval);", code_sink=self.wrapper.post_return_code)
 
-for ctype in ('char*', 'gchar*'):
+for ctype in ('char*', 'gchar*', 'const-gchar*'):
     argtypes.matcher.register_reverse_ret(ctype, StringReturn)
 del ctype
 
@@ -426,7 +428,7 @@ class VoidReturn(ReturnType):
         self.wrapper.write_code(
             code=None,
             failure_expression="py_retval != Py_None",
-            failure_cleanup='PyErr_SetString(PyExc_TypeError, "retval should be None");')
+            failure_exception='PyErr_SetString(PyExc_TypeError, "retval should be None");')
 
 argtypes.matcher.register_reverse_ret('void', VoidReturn)
 argtypes.matcher.register_reverse_ret('none', VoidReturn)
@@ -578,6 +580,21 @@ class GtkTreePathParam(IntParam):
 argtypes.matcher.register_reverse("GtkTreePath*", GtkTreePathParam)
 
 
+class GtkTreePathReturn(ReturnType):
+    def get_c_type(self):
+        return self.props.get('c_type', 'GtkTreePath *')
+    def write_decl(self):
+        self.wrapper.add_declaration("GtkTreePath * retval;")
+    def write_error_return(self):
+        self.wrapper.write_code("return NULL;")
+    def write_conversion(self):
+        self.wrapper.write_code("retval = pygtk_tree_path_from_pyobject(py_retval);\n",
+                                failure_expression=('!retval'),
+                                failure_exception=('PyErr_SetString(PyExc_TypeError, "retval should be a GtkTreePath");'))
+
+argtypes.matcher.register_reverse_ret("GtkTreePath*", GtkTreePathReturn)
+
+
 class BooleanReturn(ReturnType):
     def get_c_type(self):
         return "gboolean"
@@ -640,7 +657,7 @@ class DoubleReturn(ReturnType):
         self.wrapper.write_code(
             code=None,
             failure_expression="!PyFloat_AsDouble(py_retval)",
-            failure_cleanup='PyErr_SetString(PyExc_TypeError, "retval should be a float");')
+            failure_exception='PyErr_SetString(PyExc_TypeError, "retval should be a float");')
         self.wrapper.write_code("retval = PyFloat_AsDouble(py_retval);")
 
 for argtype in ('float', 'double', 'gfloat', 'gdouble'):
@@ -670,6 +687,7 @@ class GBoxedParam(Parameter):
 
 argtypes.matcher.register_reverse("GBoxed", GBoxedParam)
 
+
 class GBoxedReturn(ReturnType):
     def get_c_type(self):
         return self.props.get('c_type')
@@ -678,10 +696,10 @@ class GBoxedReturn(ReturnType):
     def write_error_return(self):
         self.wrapper.write_code("return retval;")
     def write_conversion(self):
-        self.wrapper.write_code(
+        self.wrapper.write_code(code = None,
             failure_expression=("!pyg_boxed_check(py_retval, %s)" %
                                 (self.props['typecode'],)),
-            failure_cleanup=('PyErr_SetString(PyExc_TypeError, "retval should be a %s");'
+            failure_exception=('PyErr_SetString(PyExc_TypeError, "retval should be a %s");'
                              % (self.props['typename'],)))
         self.wrapper.write_code('retval = pyg_boxed_get(py_retval, %s);' %
                                 self.props['typename'])
@@ -689,6 +707,52 @@ class GBoxedReturn(ReturnType):
 argtypes.matcher.register_reverse_ret("GBoxed", GBoxedReturn)
 
 
+class GdkRegionPtrReturn(GBoxedReturn):
+    def write_error_return(self):
+        self.wrapper.write_code("return gdk_region_new();")
+    def write_conversion(self):
+        self.props['typecode'] = 'PYGDK_TYPE_REGION'
+        self.props['typename'] = 'GdkRegion'
+        super(GdkRegionPtrReturn, self).write_conversion()
+
+argtypes.matcher.register_reverse_ret("GdkRegion*", GdkRegionPtrReturn)
+
+
+class PangoFontDescriptionReturn(GBoxedReturn):
+    def write_error_return(self):
+        self.wrapper.write_code("return pango_font_description_new();")
+    def write_conversion(self):
+        self.props['typecode'] = 'PANGO_TYPE_FONT_DESCRIPTION'
+        self.props['typename'] = 'PangoFontDescription'
+        super(PangoFontDescriptionReturn, self).write_conversion()
+
+argtypes.matcher.register_reverse_ret("PangoFontDescription*",
+                                      PangoFontDescriptionReturn)
+
+
+class PangoFontMetricsReturn(GBoxedReturn):
+    def write_error_return(self):
+        self.wrapper.write_code("return pango_font_metrics_new();")
+    def write_conversion(self):
+        self.props['typecode'] = 'PANGO_TYPE_FONT_METRICS'
+        self.props['typename'] = 'PangoFontMetrics'
+        super(PangoFontMetricsReturn, self).write_conversion()
+
+argtypes.matcher.register_reverse_ret("PangoFontMetrics*",
+                                      PangoFontMetricsReturn)
+
+
+class PangoLanguageReturn(GBoxedReturn):
+    def write_error_return(self):
+        self.wrapper.write_code("return pango_language_from_string(\"\");")
+    def write_conversion(self):
+        self.props['typecode'] = 'PANGO_TYPE_LANGUAGE'
+        self.props['typename'] = 'PangoLanguage'
+        super(PangoLanguageReturn, self).write_conversion()
+
+argtypes.matcher.register_reverse_ret("PangoLanguage*", PangoLanguageReturn)
+
+
 class GdkRectanglePtrParam(Parameter):
     def get_c_type(self):
         return self.props.get('c_type').replace('const-', 'const ')
